from dataclasses import dataclass


@dataclass
class V2Endpoints:
    FIND_OBJS_W_CRITERIA = "/victorwebservice/api/Objects/FindObjsWithCriteriaFilter"
    PERSIST_TO_CONTAINER = "/victorwebservice/api/Objects/PersistToContainer"
    REMOVE_FROM_CONTAINER = "/victorwebservice/api/Objects/RemoveFromContainer"
    DELETE_OBJECT = "/victorwebservice/api/Objects/Delete"
    EDIT_OBJECT = "/victorwebservice/api/Objects/Put"
    LOGIN = "/victorwebservice/api/Authenticate/Login"
    LOGOUT = "/victorwebservice/api/Authenticate/Logout"
    KEEPALIVE = "/victorwebservice/api/v2/session/keepalive"
    VERSIONS = "/victorwebservice/api/Generic/Versions"



from enum import Enum
from unittest.mock import patch

import pytest

from acslib.ccure.base import CcureACS, CcureConnection
from acslib.ccure import CcureAPI
from acslib.ccure.filters import PersonnelFilter


def test_default_ccure_acs(env_config, caplog):
    """Default picks up env vars."""
    ccure = CcureACS(connection=None)
    assert ccure.config.base_url == "https://example.com/ccure"
    assert ccure.logger.name == "acslib.ccure.connection"
    assert "acslib.ccure.connection" in caplog.text


def test_user_supplied_logger(env_config, caplog):
    """."""
    import logging

    cc_conn = CcureConnection(logger=logging.getLogger("test"))
    ccure = CcureACS(connection=cc_conn)
    assert ccure.logger.name == "test"
    assert "test:connection" in caplog.text


def test_ccure_personnel_search(env_config, personnel_response, ccure_connection, caplog):
    ccure = CcureAPI(ccure_connection)
    search_filter = PersonnelFilter(display_properties=["ObjectID", "FirstName"])
    with patch("acslib.ccure.base.CcureACS.search", return_value=personnel_response) as mock_search:
        ccure.personnel.search(terms=["test"], search_filter=search_filter)
        mock_search.assert_called_with(
            object_type="SoftwareHouse.NextGen.Common.SecurityObjects.Personnel",
            search_filter=search_filter,
            terms=["test"],
            page_size=None,
            page_number=1,
            timeout=0,
            search_options=None,
            where_clause=None,
        )
    assert "Searching for personnel" in caplog.text


@pytest.mark.skip(reason="ccure search no longer works this way")
def test_invalid_search_type(env_config):
    class NewTypes(Enum):
        NEW = "new"
        PERSONNEL = "personnel"

    ccure = CcureACS()
    with pytest.raises(ValueError):
        ccure.search(search_type=NewTypes.NEW, terms=["test"])



import random
from dataclasses import dataclass, field

import pytest
from faker import Faker

from acslib.ccure.base import CcureConnection
from acslib.ccure.config import CcureConfigFactory

fake = Faker()

BASE_CONFIG = {
    "CCURE_USERNAME": "test",
    "CCURE_PASSWORD": "test",
    "CCURE_BASE_URL": "https://example.com/ccure",
    "CCURE_CLIENT_NAME": "test",
    "CCURE_CLIENT_VERSION": "test",
    "CCURE_CLIENT_ID": "test",
}


@dataclass
class MockResponse:
    status_code: int = 200
    _json: dict = field(default_factory=dict)
    _headers: dict = field(default_factory=dict)
    text: str = ""

    def json(self):
        return self._json

    @property
    def headers(self):
        return self._headers


@pytest.fixture
def env_config(monkeypatch):
    for k, v in BASE_CONFIG.items():
        monkeypatch.setenv(k, v)


@pytest.fixture
def config():
    """."""
    return CcureConfigFactory(api_version=2, **BASE_CONFIG)


@pytest.fixture
def ccure_connection(config):
    """."""
    return CcureConnection(config=config)


@pytest.fixture
def base_mock_response():
    """."""

    def _mock_response(json=None, headers=None, text="", status_code=200):
        return MockResponse(
            **{
                "_json": json if json else {},
                "_headers": headers if headers else {},
                "text": text,
                "status_code": status_code,
            }
        )

    return _mock_response


@pytest.fixture
def response_w_session(base_mock_response):
    Faker.seed(random.randint(0, 10000))
    return base_mock_response(headers={"session-id": fake.lexify("session-?????????????????")})


@pytest.fixture
def personnel_response(response_w_session):
    response_w_session._json = {"FirstName": "Test", "MiddleName": "Ng", "LastName": "Stuff"}
    return response_w_session



"""Use CCure CRUD operations to perform some common actions"""

from datetime import datetime, timezone
from typing import Optional

from acslib.base import ACSRequestResponse
from acslib.ccure.base import CcureACS
from acslib.ccure.connection import CcureConnection
from acslib.ccure.filters import (
    CcureFilter,
    ClearanceFilter,
    PersonnelFilter,
    NFUZZ,
)
from acslib.ccure.types import ObjectType, ImageType


class PersonnelAction(CcureACS):
    def __init__(self, connection: Optional[CcureConnection] = None):
        super().__init__(connection)
        self.search_filter = PersonnelFilter()
        self.type = ObjectType.PERSONNEL.complete

    def assign_clearances(self, personnel_id: int, clearance_ids: list[int]) -> ACSRequestResponse:
        """Assign clearances to a person"""
        clearance_assignment_properties = [
            {"PersonnelID": personnel_id, "ClearanceID": clearance_id}
            for clearance_id in clearance_ids
        ]
        return self.add_children(
            parent_type=ObjectType.PERSONNEL.complete,
            parent_id=personnel_id,
            child_type=ObjectType.CLEARANCE_ASSIGNMENT.complete,
            child_configs=clearance_assignment_properties,
        )

    def revoke_clearances(self, personnel_id: int, clearance_ids: list[int]) -> ACSRequestResponse:
        """
        Revoke a person's clearances
        Two steps:
            1: Get the PersonnelClearancePair object IDs
            2: Remove those PersonnelClearancePair objects
        """

        # get PersonnelClearancePair object IDs
        clearance_query = " OR ".join(
            f"ClearanceID = {clearance_id}" for clearance_id in clearance_ids
        )
        search_filter = CcureFilter(display_properties=["PersonnelID", "ObjectID"])
        clearance_assignments = super().search(
            object_type=ObjectType.CLEARANCE_ASSIGNMENT.complete,
            search_filter=search_filter,
            terms=[personnel_id],
            page_size=0,
            where_clause=f"PersonnelID = {personnel_id} AND ({clearance_query})",
        )
        assignment_ids = [assignment.get("ObjectID") for assignment in clearance_assignments]

        if assignment_ids:
            # remove PersonnelClearancePair objects
            return self.remove_children(
                parent_type=self.type,
                parent_id=personnel_id,
                child_type=ObjectType.CLEARANCE_ASSIGNMENT.complete,
                child_ids=assignment_ids,
            )

    def get_assigned_clearances(
        self, personnel_id: int, page_size=100, page_number=1
    ) -> list[dict]:
        """Get personnel/clearance pairs associated with the given person"""
        search_filter = CcureFilter(
            lookups={"PersonnelID": NFUZZ}, display_properties=["PersonnelID", "ClearanceID"]
        )
        return super().search(
            object_type=ObjectType.CLEARANCE_ASSIGNMENT.complete,
            search_filter=search_filter,
            terms=[personnel_id],
            page_size=page_size,
            page_number=page_number,
        )

    def add_image(
        self, personnel_id: int, image: str, image_name: str = "", partition_id: int = 1
    ) -> ACSRequestResponse:
        """
        Set an image to a personnel object's PrimaryPortrait property
        - `image` is base-64 encoded.
        - `image_name` must be unique.
        - `partition_id` refers to the partition where the personnel object is stored.
        """
        if not image_name:
            timestamp = int(datetime.now(timezone.utc).timestamp())
            image_name = f"{personnel_id}_{timestamp}"
        image_properties = {
            "Name": image_name,
            "ParentId": personnel_id,
            "ImageType": ImageType.PORTRAIT.value,
            "PartitionID": partition_id,
            "Primary": True,  # this only adds primary portraits
            "Image": image,
        }
        return self.add_children(
            parent_type=ObjectType.PERSONNEL.complete,
            parent_id=personnel_id,
            child_type=ObjectType.IMAGE.complete,
            child_configs=[image_properties],
        )

    def get_image(self, personnel_id: int) -> Optional[str]:
        """
        Get the `PrimaryPortrait` property for the person with the given personnel ID.
        The returned image is a base-64 encoded string.
        """
        return self.get_property(self.type, personnel_id, "PrimaryPortrait")


class ClearanceAction(CcureACS):
    def __init__(self, connection: Optional[CcureConnection] = None):
        super().__init__(connection)
        self.search_filter = ClearanceFilter()
        self.type = ObjectType.CLEARANCE.complete

    def get_assignees(self, clearance_id: int, page_size=100, page_number=1) -> list[dict]:
        """Get clearance/personnel pairs belonging to the given clearance"""
        search_filter = CcureFilter(
            lookups={"ClearanceID": NFUZZ}, display_properties=["PersonnelID", "ClearanceID"]
        )
        return super().search(
            object_type=ObjectType.CLEARANCE_ASSIGNMENT.complete,
            search_filter=search_filter,
            terms=[clearance_id],
            page_size=page_size,
            page_number=page_number,
        )


class CcureAction:
    def __init__(self, connection: Optional[CcureConnection] = None):
        self.personnel = PersonnelAction(connection)
        self.clearance = ClearanceAction(connection)



from typing import Optional

from pydantic import BaseModel, ConfigDict


class PersonnelCreateData(BaseModel):
    """
    Validates fields for creating new Personnel objects in CCure.

    Only the LastName field is required. All other Personnel model fields are also allowed.
    """

    LastName: str

    model_config = ConfigDict(extra="allow")


class CredentialCreateData(BaseModel):
    """
    Validates fields used for creating new Credential objects in CCure,
    corresponding to fields in the ACVSCore.Credential database table.

    CHUID is required to create a new Credential. Some other common fields are suggested,
    but any Credential property is allowed.
    """

    CHUID: str
    CardNumber: Optional[int] = None
    FacilityCode: Optional[int] = None
    Name: Optional[str] = None

    model_config = ConfigDict(extra="allow")


class ClearanceItemCreateData(BaseModel):
    Name: str
    Description: str
    ParentID: int
    ParentType: str
    ControllerID: int
    ControllerClassType: str

    model_config = ConfigDict(extra="allow")



from unittest.mock import patch

import pytest
import requests

from acslib.base import ACSRequestException


def test_ccure_connection(ccure_connection, response_w_session):
    """."""
    with patch(
        "acslib.base.connection.ACSConnection._make_request", return_value=response_w_session
    ):
        session_id = ccure_connection.login()
    assert session_id == ccure_connection.session_id


@pytest.mark.parametrize(
    "side_effect",
    [
        requests.URLRequired,
        requests.HTTPError,
        requests.TooManyRedirects,
        requests.ConnectTimeout,
        requests.ReadTimeout,
        requests.Timeout,
        requests.ConnectionError,
        requests.RequestException,
    ],
)
def test_failed_login(ccure_connection, response_w_session, side_effect):
    with patch("acslib.base.connection.ACSConnection._make_request", side_effect=side_effect):
        with pytest.raises(ACSRequestException):
            ccure_connection.login()


def test_internal_server_error(ccure_connection, response_w_session):
    from acslib.base import status

    response_w_session.status_code = status.HTTP_500_INTERNAL_SERVER_ERROR
    with patch(
        "acslib.base.connection.ACSConnection._make_request", return_value=response_w_session
    ):
        with pytest.raises(ACSRequestException):
            ccure_connection.login()


def test_unhandled_request_error(ccure_connection, response_w_session):
    from acslib.base import status

    response_w_session.status_code = status.HTTP_421_MISDIRECTED_REQUEST
    with patch(
        "acslib.base.connection.ACSConnection._make_request", return_value=response_w_session
    ):
        with pytest.raises(ACSRequestException):
            ccure_connection.login()


def test_ccure_logout(ccure_connection, response_w_session):
    with patch(
        "acslib.base.connection.ACSConnection._make_request", return_value=response_w_session
    ):
        session_id = ccure_connection.login()
        assert ccure_connection._session_id == session_id
        ccure_connection.logout()
        assert ccure_connection._session_id is None



from numbers import Number
from typing import Any, Optional

from acslib.base import ACSRequestResponse
from acslib.base.connection import ACSNotImplementedException
from acslib.ccure.base import CcureACS
from acslib.ccure.connection import CcureConnection
from acslib.ccure.filters import (
    ClearanceFilter,
    ClearanceItemFilter,
    CredentialFilter,
    PersonnelFilter,
)
from acslib.ccure.data_models import (
    ClearanceItemCreateData,
    CredentialCreateData,
    PersonnelCreateData,
)
from acslib.ccure.types import ObjectType


class CcurePersonnel(CcureACS):
    def __init__(self, connection: CcureConnection):
        super().__init__(connection)
        self.search_filter = PersonnelFilter()
        self.type = ObjectType.PERSONNEL.complete

    def search(
        self,
        terms: Optional[list] = None,
        search_filter: Optional[PersonnelFilter] = None,
        page_size: Optional[int] = None,
        page_number: int = 1,
        timeout: Number = 0,
        search_options: Optional[dict] = None,
        where_clause: Optional[str] = None,
    ) -> list:
        """
        Get a list of Personnel objects matching given search terms

        :param terms: list of search terms
        :param search filter: specifies how and in what fields to look for the search terms
        """
        self.logger.info("Searching for personnel")
        search_filter = search_filter or self.search_filter

        return super().search(
            object_type=self.type,
            search_filter=search_filter,
            terms=terms,
            page_size=page_size,
            page_number=page_number,
            timeout=timeout,
            search_options=search_options,
            where_clause=where_clause,
        )

    def get_property(self, object_id: int, property_name: str) -> Any:
        return super().get_property(self.type, object_id, property_name)

    def count(
        self, terms: Optional[list] = None, search_filter: Optional[PersonnelFilter] = None
    ) -> int:
        """Get the total number of Personnel objects"""
        search_filter = search_filter or self.search_filter
        return self.search(
            search_filter=search_filter,
            terms=terms,
            search_options={"CountOnly": True},
        )

    def update(self, object_id: int, update_data: dict) -> ACSRequestResponse:
        """
        Edit properties of a personnel object

        :param object_id: the Personnel object's CCure ID
        :param update_data: maps Personnel properties to their new values
        """
        return super().update(object_type=self.type, object_id=object_id, update_data=update_data)

    def create(self, create_data: PersonnelCreateData) -> ACSRequestResponse:
        """
        Create a new personnel object

        create_data must contain a 'LastName' property.
        """
        create_data_dict = create_data.model_dump()
        property_names = list(create_data_dict)
        property_values = list(create_data_dict.values())
        request_data = {
            "Type": self.type,
            "PropertyNames": property_names,
            "PropertyValues": property_values,
        }
        return super().create(request_data=request_data)

    def delete(self, personnel_id: int) -> ACSRequestResponse:
        """Delete a personnel object by its CCure ID"""
        return super().delete(object_type=self.type, object_id=personnel_id)


class CcureClearance(CcureACS):
    def __init__(self, connection: Optional[CcureConnection] = None):
        super().__init__(connection)
        self.search_filter = ClearanceFilter()
        self.type = ObjectType.CLEARANCE.complete

    def search(
        self,
        terms: Optional[list] = None,
        search_filter: Optional[ClearanceFilter] = None,
        page_size: Optional[int] = None,
        page_number: int = 1,
        timeout: Number = 0,
        search_options: Optional[dict] = None,
        where_clause: Optional[str] = None,
    ) -> list:
        """
        Get a list of Clearance objects matching given search terms

        :param terms: list of search terms
        :param search filter: specifies how and in what fields to look for the search terms
        """
        self.logger.info("Searching for clearances")
        search_filter = search_filter or self.search_filter
        return super().search(
            object_type=self.type,
            search_filter=search_filter,
            terms=terms,
            page_size=page_size,
            page_number=page_number,
            timeout=timeout,
            search_options=search_options,
            where_clause=where_clause,
        )

    def get_property(self, object_id: int, property_name: str) -> Any:
        return super().get_property(self.type, object_id, property_name)

    def count(
        self, terms: Optional[list] = None, search_filter: Optional[ClearanceFilter] = None
    ) -> int:
        """Get the number of Clearance objects matching the search terms"""
        search_filter = search_filter or self.search_filter
        return self.search(
            search_filter=search_filter,
            terms=terms,
            search_options={"CountOnly": True},
        )

    def update(self, *args, **kwargs) -> ACSRequestResponse:
        raise ACSNotImplementedException("Updating clearances is not currently supported.")

    def create(self, *args, **kwargs) -> ACSRequestResponse:
        raise ACSNotImplementedException("Creating clearances is not currently supported.")

    def delete(self, *args, **kwargs) -> ACSRequestResponse:
        raise ACSNotImplementedException("Deleting clearances is not currently supported.")


class CcureCredential(CcureACS):
    def __init__(self, connection: Optional[CcureConnection] = None):
        super().__init__(connection)
        self.search_filter = CredentialFilter()
        self.type = ObjectType.CREDENTIAL.complete

    def search(
        self,
        terms: Optional[list] = None,
        search_filter: Optional[CredentialFilter] = None,
        page_size: Optional[int] = None,
        page_number: int = 1,
        timeout: int = 0,
        search_options: Optional[dict] = None,
        where_clause: Optional[str] = None,
    ) -> list:
        """
        Get a list of Credential objects matching given search terms

        :param terms: list of search terms
        :param search filter: specifies how and in what fields to look for the search terms
        """
        self.logger.info("Searching for credentials")
        search_filter = search_filter or self.search_filter
        return super().search(
            object_type=self.type,
            search_filter=search_filter,
            terms=terms,
            page_size=page_size,
            page_number=page_number,
            timeout=timeout,
            search_options=search_options,
            where_clause=where_clause,
        )

    def get_property(self, object_id: int, property_name: str) -> Any:
        return super().get_property(self.type, object_id, property_name)

    def count(
        self, terms: Optional[list] = None, search_filter: Optional[CredentialFilter] = None
    ) -> int:
        """Get the number of Credential objects matching the search"""
        search_filter = search_filter or self.search_filter
        return self.search(
            search_filter=search_filter,
            terms=terms,
            search_options={"CountOnly": True},
        )

    def update(self, record_id: int, update_data: dict) -> ACSRequestResponse:
        """
        Edit properties of a Credential object

        :param record_id: the Credential object's CCure ID
        :param update_data: maps Credential properties to their new values
        """
        return super().update(object_type=self.type, object_id=record_id, update_data=update_data)

    def create(self, personnel_id: int, create_data: CredentialCreateData) -> ACSRequestResponse:
        """
        Create a new credential object associated with a personnel object

        create_data properties:
            - `CHUID` is required.
            - `Name` has no effect on the new credential object.
            - `FacilityCode` defaults to 0.
            - If `CardNumber` isn't present in create_data, CHUID will be saved as 0 regardless
            of the `CHUID` value in create_data.
        """
        create_data_dict = create_data.model_dump()
        return self.add_children(
            parent_type=ObjectType.PERSONNEL.complete,
            parent_id=personnel_id,
            child_type=ObjectType.CREDENTIAL.complete,
            child_configs=[create_data_dict],
        )

    def delete(self, record_id: int) -> ACSRequestResponse:
        """Delete a Credential object by its CCure ID"""
        return super().delete(object_type=self.type, object_id=record_id)


class CcureClearanceItem(CcureACS):
    """API interactions for doors and elevators"""

    def __init__(self, connection: Optional[CcureConnection] = None):
        super().__init__(connection)
        self.search_filter = ClearanceItemFilter()

    def search(
        self,
        item_type: str,
        terms: Optional[list] = None,
        search_filter: Optional[ClearanceItemFilter] = None,
        page_size: Optional[int] = None,
        page_number: int = 1,
        timeout: Number = 0,
        search_options: Optional[dict] = None,
        where_clause: Optional[str] = None,
    ) -> list:
        """
        Get a list of ClearanceItem objects matching given search terms

        :param terms: list of search terms
        :param search filter: specifies how and in what fields to look for the search terms
        """
        self.logger.info("Searching for clearance items")
        search_filter = search_filter or self.search_filter
        return super().search(
            object_type=item_type,
            search_filter=search_filter,
            terms=terms,
            page_size=page_size,
            page_number=page_number,
            timeout=timeout,
            search_options=search_options,
            where_clause=where_clause,
        )

    def get_property(self, object_id: int, property_name: str) -> Any:
        return super().get_property(self.type, object_id, property_name)

    def count(
        self,
        item_type: str,
        terms: Optional[list] = None,
        search_filter: Optional[PersonnelFilter] = None,
    ) -> int:
        """Get the total number of ClearanceItem objects"""
        search_filter = search_filter or self.search_filter
        return self.search(
            item_type=item_type,
            search_filter=search_filter,
            terms=terms,
            search_options={"CountOnly": True},
        )

    def update(self, item_type: str, item_id: int, update_data: dict) -> ACSRequestResponse:
        """
        Edit properties of a ClearanceItem object

        :param item_type: specifies an item type. eg ObjectType.DOOR
        :param item_id: the ClearanceItem object's CCure ID
        :param update_data: maps ClearanceItem properties to their new values
        """
        return super().update(object_type=item_type, object_id=item_id, update_data=update_data)

    def create(
        self,
        item_type: str,
        controller_id: int,
        create_data: ClearanceItemCreateData,
    ) -> ACSRequestResponse:
        """
        Create a new clearance item object

        :param item_type: eg ObjectType.DOOR, ObjectType.ELEVATOR
        :param controller_id: object ID for the iStarController object for the new clearance item
        :param create_data: object with properties required to create a new clearance item
        """
        create_data_dict = create_data.model_dump()

        return self.add_children(
            parent_type=ObjectType.ISTAR_CONTROLLER,
            parent_id=controller_id,
            child_type=item_type.complete,
            child_configs=[create_data_dict],
        )

    def delete(self, item_type: str, item_id: int) -> ACSRequestResponse:
        """Delete a ClearanceItem object by its CCure ID"""
        return super().delete(object_type=item_type, object_id=item_id)



from abc import ABC

from acslib.base import ACSConnection


class AccessControlSystem(ABC):
    """Base class for all ACS implementations"""

    def __init__(self, connection: ACSConnection):
        self.connection = connection

    # Abstract ACS methods



import os

import pytest

from acslib.base import ACSConfigException
from acslib.ccure.config import CcureConfigFactory


def test_ccure_config_with_kwargs(config):
    """."""
    assert config.base_url == "https://example.com/ccure"
    assert config.connection_data == {
        "UserName": "test",
        "Password": "test",
        "ClientName": "test",
        "ClientVersion": "test",
        "ClientID": "test",
    }
    assert config.base_url == "https://example.com/ccure"
    assert str(config) == "test"
    assert config.page_size == 100
    assert config.clearance_limit == 40
    assert config.timeout == 3


def test_ccure_config_with_env_vars():
    """."""
    os.environ = {
        "CCURE_USERNAME": "test",
        "CCURE_PASSWORD": "test",
        "CCURE_BASE_URL": "https://example.com/ccure",
        "CCURE_CLIENT_NAME": "test",
        "CCURE_CLIENT_VERSION": "test",
        "CCURE_CLIENT_ID": "test",
    }
    config = CcureConfigFactory()
    assert config.connection_data == {
        "UserName": "test",
        "Password": "test",
        "ClientName": "test",
        "ClientVersion": "test",
        "ClientID": "test",
    }
    assert config.base_url == "https://example.com/ccure"
    assert str(config) == "test"
    assert config.page_size == 100
    assert config.clearance_limit == 40
    assert config.timeout == 3


def test_ccure_config_change_page_size():
    config = CcureConfigFactory(page_size=200)
    assert config.page_size == 200


def test_ccure_config_change_clearance_limit():
    config = CcureConfigFactory(clearance_limit=50)
    assert config.clearance_limit == 50


def test_ccure_config_change_timeout():
    config = CcureConfigFactory(timeout=5)
    assert config.timeout == 5


def test_no_ccure_connection_vars():
    """."""
    os.environ = {}
    with pytest.raises(ACSConfigException) as e:
        CcureConfigFactory()
    assert "Missing required environment variables" in str(e.value)


def test_missing_ccure_connection_vars():
    """."""
    os.environ = {
        "CCURE_USERNAME": "test",
        "CCURE_PASSWORD": "test",
        "CCURE_BASE_URL": "https://example.com/ccure",
        "CCURE_CLIENT_NAME": "test",
    }
    with pytest.raises(ACSConfigException) as e:
        CcureConfigFactory()
    assert "Missing required environment variables" in str(e.value)


def test_unsupported_api_version():
    """."""
    os.environ = {
        "CCURE_USERNAME": "test",
        "CCURE_PASSWORD": "test",
        "CCURE_BASE_URL": "https://example.com/ccure",
        "CCURE_CLIENT_NAME": "test",
        "CCURE_CLIENT_VERSION": "test",
        "CCURE_CLIENT_ID": "test",
    }
    with pytest.raises(ACSConfigException) as e:
        CcureConfigFactory(api_version=3)
    assert "Invalid API version: 3" in str(e.value)



from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Optional

import requests
from pydantic import BaseModel

from acslib.base import status


class ACSConnectionException(Exception):
    pass


class ACSRequestException(Exception):
    """
    Exception raised on a failed request, including exceptions
    from the requests module and 400+ status codes
    """

    def __init__(self, status_code: int, log_message: str):
        self.status_code = status_code
        self.message = log_message
        self.exception_name = "RequestException"

    def __str__(self):
        return f"{self.exception_name}: {self.status_code} {self.message}"


class ACSNotImplementedException(ACSRequestException):
    def __init__(self, log_message: str):
        self.status_code = status.HTTP_501_NOT_IMPLEMENTED
        self.message = log_message
        self.exception_name = "NotImplementedException"


class ACSRequestMethod(Enum):
    GET = "get"
    POST = "post"
    PUT = "put"
    DELETE = "delete"
    PATCH = "patch"


class ACSRequestResponse:
    """Successful queries from handle_request return this type of object"""

    def __init__(
        self, status_code: int, json: Any, headers: requests.structures.CaseInsensitiveDict
    ):
        self.status_code = status_code
        self.json = json
        self.headers = headers

    def count(self):
        return len(self.json)


class ACSRequestData(BaseModel):
    """Kwargs used in requests get/post/etc methods"""

    url: str
    # query params:
    params: Optional[dict] = None
    # body x-www-form-urlencoded data:
    data: Optional[dict | str] = None
    # body raw json:
    request_json: Optional[dict] = None
    headers: Optional[dict] = None


class ACSConnection(ABC):
    REQUEST_TYPE = {
        ACSRequestMethod.GET: requests.get,
        ACSRequestMethod.POST: requests.post,
        ACSRequestMethod.PUT: requests.put,
        ACSRequestMethod.DELETE: requests.delete,
        ACSRequestMethod.PATCH: requests.patch,
    }

    def __init__(self, **kwargs):
        self.config = kwargs.get("config")
        self.timeout = kwargs.get("timeout", 1)
        self.response = None

    @abstractmethod
    def login(self):
        pass

    @abstractmethod
    def logout(self):
        pass

    def _make_request(self, requests_method: ACSRequestMethod, request_data_map: dict):
        if req := self.REQUEST_TYPE.get(requests_method):
            return req(**request_data_map, timeout=self.timeout)
        raise ACSConnectionException(f"Invalid request method: {requests_method}")

    def request(
        self, requests_method: ACSRequestMethod, request_data: ACSRequestData
    ) -> ACSRequestResponse:
        """
        Process requests to remote servers.
        Either return a response with the resulting status code, json data, and headers,
        or raise an exception with the appropriate status code

        Parameters:
            requests_method: A method from the requests module. requests.get, requests.post, etc
            request_data: Data used as kwargs for the requests_method

        Returns: An object with status_code, json, and headers attributes
        """

        try:
            request_data_map = request_data.model_dump()
            request_data_map["json"] = request_data_map.pop("request_json", None)
            request_data_map["data"] = request_data_map.get("data", {})
            # remove request_data_map properties with None values
            request_data_map = {k: v for k, v in request_data_map.items() if v is not None}
            response = self._make_request(requests_method, request_data_map)
        except requests.HTTPError:
            # An HTTP error occurred.
            raise ACSRequestException(
                status_code=status.HTTP_400_BAD_REQUEST,
                log_message="An error occurred with this request",
            )
        except requests.URLRequired:
            # A valid URL is required to make a request.
            raise ACSRequestException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                log_message="A valid URL wasn't provided for this request",
            )
        except requests.TooManyRedirects:
            # Too many redirects.
            raise ACSRequestException(
                status_code=status.HTTP_421_MISDIRECTED_REQUEST, log_message="Too many redirects"
            )
        except requests.ConnectTimeout:
            # The request timed out while trying to connect to the remote server.
            # Requests that produced this error are safe to retry.
            raise ACSRequestException(
                status_code=status.HTTP_408_REQUEST_TIMEOUT,
                log_message=f"Unable to connect to remote server in {self.timeout} second(s)",
            )
        except requests.ReadTimeout:
            # The server did not send any data in the allotted amount of time.
            raise ACSRequestException(
                status_code=status.HTTP_408_REQUEST_TIMEOUT,
                log_message=f"No response from remote server in {self.timeout} second(s)",
            )
        except requests.Timeout:
            # The request timed out.
            # Watching this error will catch both ConnectTimeout and ReadTimeout errors.
            raise ACSRequestException(
                status_code=status.HTTP_408_REQUEST_TIMEOUT,
                log_message=f"Request took longer than {self.timeout} second(s)",
            )
        except requests.ConnectionError:
            # A Connection error occurred.
            raise ACSRequestException(
                status_code=status.HTTP_400_BAD_REQUEST,
                log_message="Could not connect to the remote host",
            )
        except requests.RequestException:
            # There was an ambiguous exception that occurred while handling your request.
            raise ACSRequestException(
                status_code=status.HTTP_400_BAD_REQUEST,
                log_message="An exception occurred while handling this request",
            )
        if response.status_code in range(200, 300):
            return ACSRequestResponse(
                status_code=response.status_code, json=response.json(), headers=response.headers
            )
        if response.status_code == status.HTTP_500_INTERNAL_SERVER_ERROR:
            raise ACSRequestException(
                status_code=status.HTTP_400_BAD_REQUEST, log_message=response.text
            )
        raise ACSRequestException(status_code=response.status_code, log_message=response.text)



"""Top-level package for Access Control Systems Library."""

__author__ = """Jeremy Gibson"""
__email__ = "jmgibso3@ncsu.edu"
__version__ = "0.1.0"

from acslib.ccure import CcureAPI
from .base.search import BooleanOperators, TermOperators



from abc import ABC, abstractmethod
from enum import Enum


class BooleanOperators(Enum):
    AND = "AND"
    OR = "OR"


class TermOperators(Enum):
    EQUALS = "="
    FUZZY = "LIKE"


class ACSFilter(ABC):
    @abstractmethod
    def filter(self, search):
        pass



import pytest

from acslib.base.search import BooleanOperators, TermOperators
from acslib.ccure.filters import (
    FUZZ,
    LFUZZ,
    NFUZZ,
    PERSONNEL_LOOKUP_FIELDS,
    CLEARANCE_LOOKUP_FIELDS,
    RFUZZ,
    ClearanceFilter,
    PersonnelFilter,
)


def test_default_instance():
    personnel_filter = PersonnelFilter()
    assert personnel_filter.filter_fields == PERSONNEL_LOOKUP_FIELDS
    assert personnel_filter.outer_bool == " AND "
    assert personnel_filter.inner_bool == " OR "
    assert personnel_filter.term_operator == "LIKE"
    assert personnel_filter.display_properties == [
        "FirstName",
        "MiddleName",
        "LastName",
        "ObjectID",
    ]

    clearance_filter = ClearanceFilter()
    assert clearance_filter.filter_fields == CLEARANCE_LOOKUP_FIELDS
    assert clearance_filter.outer_bool == " AND "
    assert clearance_filter.inner_bool == " OR "
    assert clearance_filter.term_operator == "LIKE"
    assert clearance_filter.display_properties == ["Name"]


def test_custom_instance():
    filter = PersonnelFilter(
        lookups={"Text1": NFUZZ, "Tex14": NFUZZ},
        outer_bool=BooleanOperators.OR,
        inner_bool=BooleanOperators.AND,
        term_operator=TermOperators.EQUALS,
    )
    assert filter.filter_fields == {"Text1": NFUZZ, "Tex14": NFUZZ}
    assert filter.outer_bool == " OR "
    assert filter.inner_bool == " AND "
    assert filter.term_operator == "="


def test_single_search_term():
    filter = PersonnelFilter()
    search = filter.filter(["test"])
    assert search == "(FirstName LIKE '%test%' OR LastName LIKE '%test%')"


def test_single_search_term_not_list():
    filter = PersonnelFilter()
    with pytest.raises(TypeError):
        filter.filter("test")


def test_multiple_terms():
    filter = PersonnelFilter()
    search = filter.filter(["test", "test2"])
    assert (
        search
        == "(FirstName LIKE '%test%' OR LastName LIKE '%test%') AND (FirstName LIKE '%test2%' OR LastName LIKE '%test2%')"
    )


def test_update_display_properties():
    filter = PersonnelFilter()
    filter.update_display_properties(["EmailAddress"])
    assert len(filter.display_properties) == 5
    assert "EmailAddress" in filter.display_properties


def test_update_display_properties_not_list():
    filter = PersonnelFilter()
    with pytest.raises(TypeError):
        filter.update_display_properties("EmailAddress")


def test_fuzz_functions():
    assert LFUZZ("test") == "%test"
    assert RFUZZ("test") == "test%"
    assert FUZZ("test") == "%test%"
    assert NFUZZ("test") == "test"



from numbers import Number
from typing import Optional, Any

from acslib.base import AccessControlSystem, ACSRequestData, ACSRequestResponse, ACSRequestException
from acslib.ccure.connection import CcureConnection, ACSRequestMethod
from acslib.ccure.filters import CcureFilter, NFUZZ


class CcureACS(AccessControlSystem):
    """Base class for CCure API interactions"""

    def __init__(self, connection: Optional[CcureConnection]):
        super().__init__(connection=connection)
        if not self.connection:
            self.connection = CcureConnection()
        self.logger = self.connection.logger
        self.request_options = {}

    @property
    def config(self):
        """."""
        return self.connection.config

    def search(
        self,
        object_type: str,
        terms: Optional[list] = None,
        search_filter: Optional[CcureFilter] = None,
        page_size: Optional[int] = None,
        page_number: int = 1,
        timeout: Number = 0,
        search_options: Optional[dict] = None,
        where_clause: Optional[str] = None,
    ) -> int | list:
        """
        Return CCure objects meeting the given criteria

        object_type: full name of the CCure object type, such as those in ObjectType
        search_filter: CcureFilter object specifying search query and display properties
        terms: search terms used to filter objects. Leave empty to include everything in results
        page_size: number of search results to include
                     - defaults to the page_size value in acslib/ccure/config.py
        page_number: the page of search results to display. The first page is page 1.
        search_options: other options to include in the request_json. eg. "CountOnly"
        where_clause: sql-style WHERE clause to search objects. overrides `terms` if included.
        """
        if search_filter is None and not where_clause:
            raise ACSRequestException(400, "A search filter or where clause is required.")
        if page_size is None:
            page_size = self.config.page_size
        request_json = {
            "TypeFullName": object_type,
            "pageSize": page_size,
            "pageNumber": page_number,
            "DisplayProperties": search_filter.display_properties,
            "WhereClause": where_clause or search_filter.filter(terms or []),
        } | (search_options or {})
        response = self.connection.request(
            ACSRequestMethod.POST,
            request_data=ACSRequestData(
                url=self.connection.config.base_url
                + self.connection.config.endpoints.FIND_OBJS_W_CRITERIA,
                request_json=request_json,
                headers=self.connection.base_headers,
            ),
            timeout=timeout,
        )
        return response.json

    def get_property(self, object_type: str, object_id: int, property_name: str) -> Any:
        """Return the value of one property from one CCure object"""
        search_filter = CcureFilter(lookups={"ObjectID": NFUZZ}, display_properties=[property_name])
        response = self.search(
            object_type=object_type, terms=[object_id], search_filter=search_filter, page_size=1
        )
        if response:
            search_result = response[0]
        else:
            return
        if property_name in search_result:
            return search_result[property_name]
        raise ACSRequestException(400, f"CCure object has no `{property_name}` property.")

    def update(self, object_type: str, object_id: int, update_data: dict) -> ACSRequestResponse:
        """
        Edit the properties of one CCure object

        update_data: maps property names to their new values
        """
        return self.connection.request(
            ACSRequestMethod.PUT,
            request_data=ACSRequestData(
                url=self.config.base_url + self.config.endpoints.EDIT_OBJECT,
                params={
                    "type": object_type,
                    "id": object_id,
                },
                data=self.connection.encode_data(
                    {
                        "PropertyNames": list(update_data.keys()),
                        "PropertyValues": list(update_data.values()),
                    }
                ),
                headers=self.connection.base_headers | self.connection.header_for_form_data,
            ),
        )

    def create(self, request_data: dict) -> ACSRequestResponse:
        """Persist a new CCure object"""
        return self.connection.request(
            ACSRequestMethod.POST,
            request_data=ACSRequestData(
                url=self.config.base_url + self.config.endpoints.PERSIST_TO_CONTAINER,
                data=self.connection.encode_data(request_data),
                headers=self.connection.base_headers | self.connection.header_for_form_data,
            ),
        )

    def add_children(
        self, parent_type: str, parent_id: int, child_type: str, child_configs: list[dict]
    ) -> ACSRequestResponse:
        """Persist a new CCure object as a child of an existing CCure object"""
        request_data = {
            "type": parent_type,
            "ID": parent_id,
            "Children": [
                {
                    "Type": child_type,
                    "PropertyNames": list(child_config.keys()),
                    "Propertyvalues": list(child_config.values()),
                }
                for child_config in child_configs
            ],
        }
        return self.connection.request(
            ACSRequestMethod.POST,
            request_data=ACSRequestData(
                url=self.config.base_url + self.config.endpoints.PERSIST_TO_CONTAINER,
                data=self.connection.encode_data(request_data),
                headers=self.connection.base_headers | self.connection.header_for_form_data,
            ),
        )

    def remove_children(
        self, parent_type: str, parent_id: int, child_type: str, child_ids: list[int]
    ) -> ACSRequestResponse:
        """Remove child CCure objects from a parent CCure object"""
        request_data = {
            "type": parent_type,
            "ID": parent_id,
            "Children": [
                {
                    "Type": child_type,
                    "ID": child_id,
                }
                for child_id in child_ids
            ],
        }
        return self.connection.request(
            ACSRequestMethod.POST,
            request_data=ACSRequestData(
                url=self.config.base_url + self.config.endpoints.REMOVE_FROM_CONTAINER,
                data=self.connection.encode_data(request_data),
                headers=self.connection.base_headers | self.connection.header_for_form_data,
            ),
        )

    def delete(self, object_type: str, object_id: int) -> ACSRequestResponse:
        """Delete a CCure object"""
        return self.connection.request(
            ACSRequestMethod.DELETE,
            request_data=ACSRequestData(
                url=self.config.base_url + self.config.endpoints.DELETE_OBJECT,
                params={"type": object_type, "id": object_id},
                headers=self.connection.base_headers,
            ),
        )



from typing import Optional

from acslib.base.search import ACSFilter, BooleanOperators, TermOperators


def left_fuzz(term):
    return f"%{term}"


def right_fuzz(term):
    return f"{term}%"


def full_fuzz(term):
    return f"%{term}%"


def no_fuzz(term):
    return f"{term}"


LFUZZ = left_fuzz
RFUZZ = right_fuzz
FUZZ = full_fuzz
NFUZZ = no_fuzz

PERSONNEL_LOOKUP_FIELDS = {"FirstName": FUZZ, "LastName": FUZZ}
CLEARANCE_LOOKUP_FIELDS = {"Name": FUZZ}
CREDENTIAL_LOOKUP_FIELDS = {"Name": FUZZ}
CLEARANCE_ITEM_LOOKUP_FIELDS = {"Name": FUZZ}


class CcureFilter(ACSFilter):
    """Base CCure Filter
    :param lookups: Dict containing searchable field names and their lookup functions
    :param outer_bool: Boolean operator to use between search terms
    :param inner_bool: Boolean operator to use between lookups
    :param term_operator: Term operator to use between field and a search term
    :attribute
    """

    def __init__(
        self,
        lookups: dict[str, callable] = {"ObjectID": NFUZZ},
        outer_bool=BooleanOperators.AND,
        inner_bool=BooleanOperators.OR,
        term_operator=TermOperators.FUZZY,
        display_properties: list[str] = [],
    ):
        self.filter_fields = lookups
        self.outer_bool = outer_bool.value
        self.inner_bool = inner_bool.value
        self.term_operator = term_operator.value
        #: List of properties from CCURE to be included in the CCURE response
        self.display_properties = display_properties

    def _compile_term(self, term) -> str:
        """Get all parts of the query for one search term"""
        fields = [(field_name, lookup(term)) for field_name, lookup in self.filter_fields.items()]
        field_queries = [
            f"{field_name} {self.term_operator} '{lookup}'" for field_name, lookup in fields
        ]
        return f"({self.inner_bool.join(field_queries)})"

    def update_display_properties(self, properties: list[str]):
        if not isinstance(properties, list):
            raise TypeError("Properties must be a list of strings")
        self.display_properties += properties

    def filter(self, search: list[str]) -> str:
        if not isinstance(search, list):
            raise TypeError("Search must be a list of strings")
        return self.outer_bool.join(self._compile_term(term) for term in search)


class PersonnelFilter(CcureFilter):
    """Basic CCure Personnel Filter
    :param lookups: Dict containing searchable field names and their lookup functions
    :param outer_bool: Boolean operator to use between search terms
    :param inner_bool: Boolean operator to use between lookups
    :param term_operator: Term operator to use between field and a search term
    :param display_properties: List of properties from CCure to be included in the CCure response
    :attribute
    """

    def __init__(
        self,
        lookups: Optional[dict[str, callable]] = None,
        outer_bool=BooleanOperators.AND,
        inner_bool=BooleanOperators.OR,
        term_operator=TermOperators.FUZZY,
        display_properties: Optional[list[str]] = None,
    ):
        self.filter_fields = lookups or PERSONNEL_LOOKUP_FIELDS
        self.outer_bool = f" {outer_bool.value} "
        self.inner_bool = f" {inner_bool.value} "
        self.term_operator = term_operator.value
        self.display_properties = ["FirstName", "MiddleName", "LastName", "ObjectID"]
        if display_properties is not None:
            self.display_properties = display_properties


class ClearanceFilter(CcureFilter):
    """Basic CCure Clearance Filter
    :param lookups: Dict containing searchable field names and their lookup functions
    :param outer_bool: Boolean operator to use between search terms
    :param inner_bool: Boolean operator to use between lookups
    :param term_operator: Term operator to use between field and a search term
    :param display_properties: List of properties from CCure to be included in the CCure response
    :attribute
    """

    def __init__(
        self,
        lookups: Optional[dict[str, callable]] = None,
        outer_bool=BooleanOperators.AND,
        inner_bool=BooleanOperators.OR,
        term_operator=TermOperators.FUZZY,
        display_properties: Optional[list[str]] = None,
    ):
        self.filter_fields = lookups or CLEARANCE_LOOKUP_FIELDS
        self.outer_bool = f" {outer_bool.value} "
        self.inner_bool = f" {inner_bool.value} "
        self.term_operator = term_operator.value
        self.display_properties = ["Name"]
        if display_properties is not None:
            self.display_properties = display_properties


class CredentialFilter(CcureFilter):
    """Basic CCure Credential Filter
    :param lookups: Dict containing searchable field names and their lookup functions
    :param outer_bool: Boolean operator to use between search terms
    :param inner_bool: Boolean operator to use between lookups
    :param term_operator: Term operator to use between field and a search term
    :param display_properties: List of properties from CCure to be included in the CCure response
    :attribute
    """

    def __init__(
        self,
        lookups: Optional[dict[str, callable]] = None,
        outer_bool=BooleanOperators.AND,
        inner_bool=BooleanOperators.OR,
        term_operator=TermOperators.FUZZY,
        display_properties: Optional[list[str]] = None,
    ):
        self.filter_fields = lookups or CREDENTIAL_LOOKUP_FIELDS
        self.outer_bool = f" {outer_bool.value} "
        self.inner_bool = f" {inner_bool.value} "
        self.term_operator = term_operator.value
        self.display_properties = ["Name"]
        if display_properties is not None:
            self.display_properties = display_properties


class ClearanceItemFilter(CcureFilter):
    """Basic CCure ClearanceItem Filter
    :param lookups: Dict containing searchable field names and their lookup functions
    :param outer_bool: Boolean operator to use between search terms
    :param inner_bool: Boolean operator to use between lookups
    :param term_operator: Term operator to use between field and a search term
    :param display_properties: List of properties from CCure to be included in the CCure response
    :attribute
    """

    def __init__(
        self,
        lookups: Optional[dict[str, callable]] = None,
        outer_bool=BooleanOperators.AND,
        inner_bool=BooleanOperators.OR,
        term_operator=TermOperators.FUZZY,
        display_properties: Optional[list[str]] = None,
    ):
        self.filter_fields = lookups or CLEARANCE_ITEM_LOOKUP_FIELDS
        self.outer_bool = f" {outer_bool.value} "
        self.inner_bool = f" {inner_bool.value} "
        self.term_operator = term_operator.value
        self.display_properties = ["Name"]
        if display_properties is not None:
            self.display_properties = display_properties



from .config import ACSConfig, ACSConfigException
from .connection import ACSConnection, ACSRequestData, ACSRequestException, ACSRequestResponse
from .acs import AccessControlSystem
import acslib.base.status as status



from enum import Enum


class ObjectType(Enum):
    CLEARANCE = "clearance"
    CLEARANCE_ASSIGNMENT = "clearance assignment"
    CREDENTIAL = "credential"
    CLEARANCE_ITEM = "clearance item"
    DOOR = "door"
    ELEVATOR = "elevator"
    IMAGE = "image"
    ISTAR_CONTROLLER = "istar controller"
    PERSONNEL = "personnel"
    TIME_SPEC = "time spec"

    @property
    def complete(self):
        if self == self.CLEARANCE:
            return "SoftwareHouse.NextGen.Common.SecurityObjects.Clearance"
        if self == self.CLEARANCE_ASSIGNMENT:
            return "SoftwareHouse.NextGen.Common.SecurityObjects.PersonnelClearancePair"
        if self == self.CREDENTIAL:
            return "SoftwareHouse.NextGen.Common.SecurityObjects.Credential"
        if self == self.CLEARANCE_ITEM:
            return "SoftwareHouse.NextGen.Common.SecurityObjects.ClearanceItem"
        if self == self.DOOR:
            return "SoftwareHouse.NextGen.Common.SecurityObjects.Door"
        if self == self.ELEVATOR:
            return "SoftwareHouse.NextGen.Common.SecurityObjects.Elevator"
        if self == self.IMAGE:
            return "SoftwareHouse.NextGen.Common.SecurityObjects.Images"
        if self == self.ISTAR_CONTROLLER:
            return "SoftwareHouse.NextGen.Common.SecurityObjects.iStarController"
        if self == self.PERSONNEL:
            return "SoftwareHouse.NextGen.Common.SecurityObjects.Personnel"
        if self == self.TIME_SPEC:
            return "SoftwareHouse.CrossFire.Common.Objects.TimeSpec"


class ImageType(Enum):
    UNKNOWN = 0
    PORTRAIT = 1
    SIGNATURE = 2
    FINGERPRINT = 3
    HANDPRINT = 4
    DYNAMIC_BADGE_IMAGE = 5
    STATIC_BADGE_IMAGE = 6
    SYSTEM_IMAGE = 7
    PRIVATE_IMAGE = 8
    SHARED_IMAGE = 9



import logging
from numbers import Number
from typing import Optional

from acslib.base import (
    ACSConnection,
    ACSRequestData,
    ACSRequestException,
    ACSRequestResponse,
    status,
)
from acslib.base.connection import ACSRequestMethod
from acslib.ccure.config import CcureConfigFactory


class CcureConnection(ACSConnection):
    def __init__(self, **kwargs):
        """
        A connection object to the CCure Server.
        Parameters:
        :param kwargs:
        """
        self._session_id = None
        if conn_logger := kwargs.get("logger"):
            self.logger = conn_logger
        else:
            self.logger = logging.getLogger(__name__)

        if not kwargs.get("config"):
            kwargs["config"] = CcureConfigFactory()
        self.logger.info("Initializing CCure connection")
        super().__init__(**kwargs)

    @property
    def session_id(self) -> str:
        if self._session_id:
            return self._session_id
        return self.login()

    @property
    def base_headers(self):
        """Headers required for each request to CCure"""
        return {
            "session-id": self.session_id,
            "Access-Control-Expose-Headers": "session-id",
        }

    @property
    def header_for_form_data(self):
        return {"Content-Type": "application/x-www-form-urlencoded"}

    def login(self):
        """Open a new CCure session and generate a new session ID"""
        try:
            response = self.request(
                ACSRequestMethod.POST,
                request_data=ACSRequestData(
                    url=self.config.base_url + self.config.endpoints.LOGIN,
                    data=self.config.connection_data,
                ),
            )
            self._session_id = response.headers["session-id"]
            self.logger.debug(f"Fetched new Session ID: {self._session_id}")
        except ACSRequestException as e:
            self.logger.error(f"Error Fetching Session ID: {e}")
            self.log_session_details()
            self.logger.debug(f"Connection data: {self.config.connection_data}")
            raise e
        return self._session_id

    def logout(self):
        """Log out of the CCure session"""
        if self._session_id:
            self.logger.debug(f"Logging out of CCure session: {self._session_id}")
            try:
                self.request(
                    ACSRequestMethod.POST,
                    request_data=ACSRequestData(
                        url=self.config.base_url + self.config.endpoints.LOGOUT,
                        headers={"session-id": self._session_id},
                    ),
                )
            except ACSRequestException as e:
                self.logger.error(f"Error logging out of CCure session: {e}")
                self.log_session_details()
            finally:
                self.logger.debug(f"Removing Session ID: {self._session_id}")
                self._session_id = None

    def keepalive(self):
        """Prevent the CCure api session from expiring from inactivity"""
        self.logger.debug(f"Keeeping CCure session alive: {self.session_id}")
        try:
            self.request(
                ACSRequestMethod.POST,
                request_data=ACSRequestData(
                    url=self.config.base_url + self.config.endpoints.KEEPALIVE,
                    headers={
                        "session-id": self.session_id,
                        "Access-Control-Expose-Headers": "session-id",
                    },
                ),
            )
            self.logger.debug(f"Session kept alive: {self.session_id}")
        except ACSRequestException as e:
            self.logger.error(f"Error keeping CCure session alive: {e}")
            self.log_session_details()
            self.logout()

    def request(
        self,
        requests_method: ACSRequestMethod,
        request_data: ACSRequestData,
        timeout: Optional[Number] = 0,
        request_attempts: int = 2,
    ) -> ACSRequestResponse:
        """
        Call the `ACSConnection.handle_requests` function and return the result.
        If the response is a 401, get a new CCure session_id and try the request again.

        Parameters:
            requests_method: A method from the requests module. get, post, etc
            request_data: Data used as kwargs for the requests_method
            timeout: Maximum time to wait for a server response, in seconds
            request_attempts: Maximum number of times to try the request

        Returns: An object with status_code, json, and headers attributes
        """
        # use TIMEOUT as the default timeout value
        if timeout != 0:
            self.config.timeout = timeout
        while request_attempts > 0:
            try:
                return super().request(requests_method, request_data)
            except ACSRequestException as e:
                if e.status_code != status.HTTP_401_UNAUTHORIZED or request_attempts == 1:
                    raise e
                request_attempts -= 1
                self.logout()
                request_data.headers["session-id"] = self.session_id

    def log_session_details(self):
        """Log session ID and the api version number"""
        version_url = self.config.base_url + self.config.endpoints.VERSIONS
        self.logger.error(f"Session ID: {self._session_id}")
        try:
            response = self.request(
                ACSRequestMethod.POST,
                request_data=ACSRequestData(url=version_url),
            ).json
            self.logger.debug(f"CCure webservice version: {response.get('webServiceVersion')}")
            self.logger.debug(f"CCure app server version: {response.get('appServerVersion')}")
        except ACSRequestException as e:
            self.logger.debug(f"Could not get CCure api version number: {e}")

    @staticmethod
    def encode_data(data: dict) -> str:
        """
        Encode a dictionary of form data as a string for requests

        Parameters:
            data: form data for the request

        Returns: the string of encoded data
        """

        def get_form_entries(data: dict, prefix: str = "") -> list[str]:
            """
            Convert the data dict into a list of form entries

            Parameters:
                data: data about the new clearance assignment

            Returns: list of strings representing key/value pairs
            """
            entries = []
            for key, val in data.items():
                if isinstance(val, (int, float, str)):
                    if prefix:
                        entries.append(f"{prefix}[{key}]={val}")
                    else:
                        entries.append(f"{key}={val}")
                elif isinstance(val, (list, tuple)):
                    for i, list_item in enumerate(val):
                        if isinstance(list_item, dict):
                            entries.extend(
                                get_form_entries(data=list_item, prefix=prefix + f"{key}[{i}]")
                            )
                        elif prefix:
                            entries.append(f"{prefix}[{key}][]={list_item}")
                        else:
                            entries.append(f"{key}[]={list_item}")
            return entries

        return "&".join(get_form_entries(data))



from typing import Optional

from acslib.base import ACSRequestResponse
from acslib.base.connection import ACSRequestData, ACSRequestMethod
from acslib.ccure.actions import CcureAction
from acslib.ccure.base import CcureACS
from acslib.ccure.connection import CcureConnection
from acslib.ccure.crud import (
    CcurePersonnel,
    CcureClearance,
    CcureCredential,
    CcureClearanceItem,
)
from acslib.ccure.filters import ClearanceFilter, PersonnelFilter, CredentialFilter


class CcureAPI:
    def __init__(self, connection: Optional[CcureConnection] = None):
        self.connection = connection or CcureConnection()
        self.personnel = CcurePersonnel(self.connection)
        self.clearance = CcureClearance(self.connection)
        self.credential = CcureCredential(self.connection)
        self.clearance_item = CcureClearanceItem(self.connection)
        self.action = CcureAction(self.connection)
        self.ccure_object = CcureACS(self.connection)



import os

from acslib.base.config import ACSConfig, ACSConfigException
from acslib.ccure.endpoints import V2Endpoints


class CcureConfig(ACSConfig):
    """
    CcureConfig returns an object that implements the ACSConfig interface.
    It expects the following variables to be set either in the environment or passed as kwargs:
        CCURE_USERNAME
        CCURE_PASSWORD
        CCURE_BASE_URL
        CCURE_CLIENT_NAME
        CCURE_CLIENT_VERSION
        CCURE_CLIENT_ID
    :param PAGE_SIZE: default 100
    :param CLEARANCE_LIMIT: default 40
    :param TIMEOUT: default 3
    :param kwargs:
    :return: CcureConfig
    """

    def __init__(self, **kwargs):
        self.page_size = kwargs.get("page_size", 100)
        self.current_page = 1
        self.clearance_limit = kwargs.get("clearance_limit", 40)
        self.timeout = kwargs.get("timeout", 3)
        self.endpoints = None
        self.username = kwargs.get("CCURE_USERNAME", os.getenv("CCURE_USERNAME"))
        self.password = kwargs.get("CCURE_PASSWORD", os.getenv("CCURE_PASSWORD"))
        self.base_url = kwargs.get("CCURE_BASE_URL", os.getenv("CCURE_BASE_URL"))
        self.client_name = kwargs.get("CCURE_CLIENT_NAME", os.getenv("CCURE_CLIENT_NAME"))
        self.client_version = kwargs.get("CCURE_CLIENT_VERSION", os.getenv("CCURE_CLIENT_VERSION"))
        self.client_id = kwargs.get("CCURE_CLIENT_ID", os.getenv("CCURE_CLIENT_ID"))
        if not all(
            [
                self.username,
                self.password,
                self.base_url,
                self.client_name,
                self.client_version,
                self.client_id,
            ]
        ):
            raise ACSConfigException(
                "Missing required environment variables: "
                "CCURE_USERNAME, CCURE_PASSWORD, CCURE_BASE_URL, CCURE_CLIENT_NAME, "
                "CCURE_CLIENT_VERSION, CCURE_CLIENT_ID"
            )

    def __str__(self):
        return self.client_version

    @property
    def connection_data(self) -> dict:
        return {
            "UserName": self.username,
            "Password": self.password,
            "ClientName": self.client_name,
            "ClientVersion": self.client_version,
            "ClientID": self.client_id,
        }


class CcureConfigFactory:
    def __new__(cls, *args, **kwargs) -> CcureConfig:
        """
        CcuureConfigFactory returns a CcureConfig instance with the correct endpoints for the requested API version.
        The default api version is 2.
        :param args:
        :param kwargs:
        :return: CcureConfig
        """
        api_version = kwargs.get("api_version", 2)
        instance = CcureConfig(**kwargs)
        match api_version:
            case 2:
                instance.endpoints = V2Endpoints
            case _:
                raise ACSConfigException(f"Invalid API version: {api_version}")
        return instance
